@@ -1,4 +1,9 @@
 from __future__ import print_function
+from imblearn.under_sampling import RandomUnderSampler
+from imblearn.over_sampling import SMOTE
+from sklearn.model_selection import GroupShuffleSplit
+from pyAudioAnalysis import audioBasicIO
+from pyAudioAnalysis import MidTermFeatures as aF
 import sys
 import numpy as np
 import os
@@ -19,9 +24,6 @@ from sklearn.model_selection import train_test_split
 from sklearn.preprocessing import StandardScaler
 sys.path.insert(0, os.path.join(
     os.path.dirname(os.path.realpath(__file__)), "../"))
-from pyAudioAnalysis import MidTermFeatures as aF
-from pyAudioAnalysis import audioBasicIO
-from sklearn.model_selection import GroupShuffleSplit
 
 shortTermWindow = 0.050
 shortTermStep = 0.050
@@ -43,7 +45,7 @@ class Knn:
         P = np.zeros((n_classes,))
         for i in range(n_classes):
             P[i] = np.nonzero(self.labels[i_sort[0]
-                [0:self.neighbors]] == i)[0].shape[0] / float(self.neighbors)
+                                          [0:self.neighbors]] == i)[0].shape[0] / float(self.neighbors)
         return np.argmax(P), P
 
 
@@ -106,7 +108,7 @@ def regression_wrapper(model, model_type, test_sample):
     """
     if model_type == "svm" or model_type == "randomforest" or \
             model_type == "svm_rbf":
-        return model.predict(test_sample.reshape(1,-1))[0]
+        return model.predict(test_sample.reshape(1, -1))[0]
 
     #    elif classifier_type == "knn":
     #    TODO
@@ -233,7 +235,9 @@ def train_random_forest_regression(features, labels, n_estimators):
 
 def extract_features_and_train(paths, mid_window, mid_step, short_window,
                                short_step, classifier_type, model_name,
-                               compute_beat=False, train_percentage=0.90, dict_of_ids=None):
+                               compute_beat=False, train_percentage=0.90,
+                               dict_of_ids=None,
+                               use_smote=False):
     """
     This function is used as a wrapper to segment-based audio feature extraction
     and classifier training.
@@ -281,7 +285,7 @@ def extract_features_and_train(paths, mid_window, mid_step, short_window,
     elif classifier_type == "randomforest":
         classifier_par = np.array([10, 25, 50, 100, 200, 500])
     elif classifier_type == "knn":
-        classifier_par = np.array([1, 3, 5, 7, 9, 11, 13, 15])        
+        classifier_par = np.array([1, 3, 5, 7, 9, 11, 13, 15])
     elif classifier_type == "gradientboosting":
         classifier_par = np.array([10, 25, 50, 100, 200, 500])
     elif classifier_type == "extratrees":
@@ -301,16 +305,21 @@ def extract_features_and_train(paths, mid_window, mid_step, short_window,
     features = temp_features
 
     best_param = evaluate_classifier(features, class_names, classifier_type,
-                                     classifier_par, 0, list_of_ids, n_exp=-1,
-                                     train_percentage=train_percentage)
+                                     classifier_par, 1, list_of_ids, n_exp=-1,
+                                     train_percentage=train_percentage,
+                                     smote=use_smote)
 
     print("Selected params: {0:.5f}".format(best_param))
 
-
     # STEP C: Train and Save the classifier to file
-
-    # First Use mean/std standard feature scaling:
+    # Get featues in the X, y format:
     features, labels = features_to_matrix(features)
+    # Apply smote if necessary:
+    if use_smote:
+        sm = SMOTE(random_state=2)
+        features, labels = sm.fit_resample(features, labels)
+   
+    # Use mean/std standard feature scaling:
     scaler = StandardScaler()
     features = scaler.fit_transform(features)
     mean = scaler.mean_.tolist()
@@ -328,7 +337,7 @@ def extract_features_and_train(paths, mid_window, mid_step, short_window,
     elif classifier_type == "extratrees":
         classifier = train_extra_trees(features, labels, best_param)
 
-    # And save the model to a file, along with 
+    # And save the model to a file, along with
     # - the scaling -mean/std- vectors)
     # - the feature extraction parameters
     if classifier_type == "knn":
@@ -434,9 +443,9 @@ def feature_extraction_train_regression(folder_name, mid_window, mid_step,
         print("Regression task " + r)
         bestParam, error, berror = evaluate_regression(f_final[iRegression],
                                                        regression_labels[
-                                                          iRegression],
-                                                       100, model_type,
-                                                       model_params)
+            iRegression],
+            100, model_type,
+            model_params)
         errors.append(error)
         errors_base.append(berror)
         best_params.append(bestParam)
@@ -461,10 +470,10 @@ def feature_extraction_train_regression(folder_name, mid_window, mid_step,
         if model_type == "randomforest":
             classifier, _ = train_random_forest_regression(features_norm,
                                                            regression_labels[
-                                                            iRegression],
+                                                               iRegression],
                                                            bestParam)
 
-        # Save the model to a file, along with 
+        # Save the model to a file, along with
         # - the scaling -mean/std- vectors)
         # - the feature extraction parameters
         if model_type == "svm" or model_type == "svm_rbf" \
@@ -503,10 +512,10 @@ def load_model_knn(knn_model_name, is_regression=False):
 
     if is_regression:
         return classifier, mean, std, mid_window, mid_step, short_window, \
-               short_step, compute_beat
+            short_step, compute_beat
     else:
         return classifier, mean, std, classes, mid_window, mid_step, \
-               short_window, short_step, compute_beat
+            short_window, short_step, compute_beat
 
 
 def load_model(model_name, is_regression=False):
@@ -536,10 +545,11 @@ def load_model(model_name, is_regression=False):
 
     if is_regression:
         return svm_model, mean, std, mid_window, mid_step, short_window, \
-               short_step, compute_beat
+            short_step, compute_beat
     else:
         return svm_model, mean, std, classNames, mid_window, mid_step, \
-               short_window, short_step, compute_beat
+            short_window, short_step, compute_beat
+
 
 def group_split(X, y, train_indeces, test_indeces, split_id):
     """
@@ -560,8 +570,11 @@ def group_split(X, y, train_indeces, test_indeces, split_id):
     y_train, y_test = y[train_index], y[test_index]
     return X_train, X_test, y_train, y_test
 
+
 def evaluate_classifier(features, class_names, classifier_name, params,
-                        parameter_mode, list_of_ids=None, n_exp=-1, train_percentage=0.90):
+                        parameter_mode, list_of_ids=None, n_exp=-1,
+                        train_percentage=0.90,
+                        smote=False):
     """
     ARGUMENTS:
         features:     a list ([numOfClasses x 1]) whose elements containt
@@ -592,6 +605,7 @@ def evaluate_classifier(features, class_names, classifier_name, params,
     n_classes = len(features)
     ac_all = []
     f1_all = []
+    f1_std_all = []
     pre_class_all = []
     rec_classes_all = []
     f1_classes_all = []
@@ -603,7 +617,7 @@ def evaluate_classifier(features, class_names, classifier_name, params,
     n_samples_total = X.shape[0]
 
     if n_exp == -1:
-        n_exp = int(10000 / n_samples_total) + 1
+        n_exp = int(50000 / n_samples_total) + 1
 
     if list_of_ids:
         train_indeces, test_indeces = [], []
@@ -612,25 +626,32 @@ def evaluate_classifier(features, class_names, classifier_name, params,
             train_indeces.append(train_index)
             test_indeces.append(test_index)
 
-
     for Ci, C in enumerate(params):
         # for each param value
         cm = np.zeros((n_classes, n_classes))
+        f1_per_exp = []
+        y_pred_all = []
+        y_test_all = []
         for e in range(n_exp):
-            y_pred, y_real = [], []
+            y_pred = []
             # for each cross-validation iteration:
             print("Param = {0:.5f} - classifier Evaluation "
                   "Experiment {1:d} of {2:d}".format(C, e+1, n_exp))
             # split features:
 
             if list_of_ids:
-                X_train, X_test, y_train, y_test = group_split(X, y, train_indeces, test_indeces, e)
+                X_train, X_test, y_train, y_test = group_split(
+                    X, y, train_indeces, test_indeces, e)
             else:
                 X_train, X_test, y_train, y_test = \
                     train_test_split(X, y, test_size=1-train_percentage)
 
             # mean/std scale the features:
             scaler = StandardScaler()
+            if smote:
+                sm = SMOTE(random_state=2)
+                #sm = RandomUnderSampler(random_state=0)
+                X_train, y_train = sm.fit_resample(X_train, y_train)
             scaler.fit(X_train)
             X_train = scaler.transform(X_train)
 
@@ -646,17 +667,24 @@ def evaluate_classifier(features, class_names, classifier_name, params,
             elif classifier_name == "gradientboosting":
                 classifier = train_gradient_boosting(X_train, y_train, C)
             elif classifier_name == "extratrees":
-                classifier = train_extra_trees(X_train, y_train,C)
+                classifier = train_extra_trees(X_train, y_train, C)
 
             # get predictions and compute current comfusion matrix
             cmt = np.zeros((n_classes, n_classes))
             X_test = scaler.transform(X_test)
             for i_test_sample in range(X_test.shape[0]):
                 y_pred.append(classifier_wrapper(classifier,
-                                                 classifier_name, 
+                                                 classifier_name,
                                                  X_test[i_test_sample, :])[0])
+            # current confusion matrices and F1:
             cmt = sklearn.metrics.confusion_matrix(y_test, y_pred)
+            f1t = sklearn.metrics.f1_score(y_test, y_pred, average='macro')
+            # aggregated predicted and ground truth labels 
+            # (used for the validation of final F1)
+            y_pred_all += y_pred
+            y_test_all += y_test.tolist()
 
+            f1_per_exp.append(f1t)
             if cmt.size != cm.size:
                 all_classes = set(y)
                 split_classes = set(y_test)
@@ -668,20 +696,31 @@ def evaluate_classifier(features, class_names, classifier_name, params,
             cm = cm + cmt
         cm = cm + 0.0000000010
 
-        rec = np.array([cm[ci, ci] / np.sum(cm[ci, :]) 
-                        for ci in range(cm.shape[0])]) 
-        pre = np.array([cm[ci, ci] / np.sum(cm[:, ci]) 
+        rec = np.array([cm[ci, ci] / np.sum(cm[ci, :])
+                        for ci in range(cm.shape[0])])
+        pre = np.array([cm[ci, ci] / np.sum(cm[:, ci])
                         for ci in range(cm.shape[0])])
 
         pre_class_all.append(pre)
         rec_classes_all.append(rec)
 
         f1 = 2 * rec * pre / (rec + pre)
+
+        # this is just for debugging (it should be equal to f1)
+        f1_b = sklearn.metrics.f1_score(y_test_all, y_pred_all,
+                                        average='macro')
+        # Note: np.mean(f1_per_exp) will not be exacty equal to the
+        # overall f1 (i.e. f1 and f1_b because these are calculated on a
+        # per-sample basis)
+        f1_std = np.std(f1_per_exp)
+        #print(np.mean(f1), f1_b, f1_std)
+
         f1_classes_all.append(f1)
         ac_all.append(np.sum(np.diagonal(cm)) / np.sum(cm))
 
         cms_all.append(cm)
         f1_all.append(np.mean(f1))
+        f1_std_all.append(f1_std)
 
     print("\t\t", end="")
     for i, c in enumerate(class_names):
@@ -723,6 +762,8 @@ def evaluate_classifier(features, class_names, classifier_name, params,
         # keep parameters that maximize overall f1 measure:
         print("Confusion Matrix:")
         print_confusion_matrix(cms_all[best_f1_ind], class_names)
+        print(f"Best macro f1 {100 * f1_all[best_f1_ind]:.1f}")
+        print(f"Best macro f1 std {100 * f1_std_all[best_f1_ind]:.1f}")
         return params[best_f1_ind]
 
 
@@ -749,50 +790,50 @@ def evaluate_regression(features, labels, n_exp, method_name, params):
     er_train_all = []
     er_base_all = []
     for Ci, C in enumerate(params):   # for each param value
-                errors = []
-                errors_train = []
-                errors_baseline = []
-                for e in range(n_exp):   # for each cross-validation iteration:
-                    # split features:
-                    randperm = np.random.permutation(range(n_samples))
-                    n_train = int(round(per_train * n_samples))
-                    f_train = [features_norm[randperm[i]]
-                               for i in range(n_train)]
-                    f_test = [features_norm[randperm[i+n_train]]
-                              for i in range(n_samples - n_train)]
-                    l_train = [labels[randperm[i]] for i in range(n_train)]
-                    l_test = [labels[randperm[i + n_train]]
-                              for i in range(n_samples - n_train)]
+        errors = []
+        errors_train = []
+        errors_baseline = []
+        for e in range(n_exp):   # for each cross-validation iteration:
+            # split features:
+            randperm = np.random.permutation(range(n_samples))
+            n_train = int(round(per_train * n_samples))
+            f_train = [features_norm[randperm[i]]
+                       for i in range(n_train)]
+            f_test = [features_norm[randperm[i+n_train]]
+                      for i in range(n_samples - n_train)]
+            l_train = [labels[randperm[i]] for i in range(n_train)]
+            l_test = [labels[randperm[i + n_train]]
+                      for i in range(n_samples - n_train)]
 
-                    # train multi-class svms:                    
-                    f_train = np.matrix(f_train)                                 
-                    if method_name == "svm":                                        
-                        classifier, train_err = \
-                            train_svm_regression(f_train, l_train, C)
-                    elif method_name == "svm_rbf":                      
-                        classifier, train_err = \
-                            train_svm_regression(f_train, l_train, C,
-                                                 kernel='rbf')
-                    elif method_name == "randomforest":
-                        classifier, train_err = \
-                            train_random_forest_regression(f_train, l_train, C)
-                    error_test = []
-                    error_test_baseline = []
-                    for itest, fTest in enumerate(f_test):
-                        R = regression_wrapper(classifier, method_name, fTest)
-                        Rbaseline = np.mean(l_train)
-                        error_test.append((R - l_test[itest]) *
-                                          (R - l_test[itest]))
-                        error_test_baseline.append((Rbaseline - l_test[itest]) *
-                                                  (Rbaseline - l_test[itest]))
-                    error = np.array(error_test).mean()
-                    error_baseline = np.array(error_test_baseline).mean()
-                    errors.append(error)
-                    errors_train.append(train_err)
-                    errors_baseline.append(error_baseline)
-                errors_all.append(np.array(errors).mean())
-                er_train_all.append(np.array(errors_train).mean())
-                er_base_all.append(np.array(errors_baseline).mean())
+            # train multi-class svms:
+            f_train = np.matrix(f_train)
+            if method_name == "svm":
+                classifier, train_err = \
+                    train_svm_regression(f_train, l_train, C)
+            elif method_name == "svm_rbf":
+                classifier, train_err = \
+                    train_svm_regression(f_train, l_train, C,
+                                         kernel='rbf')
+            elif method_name == "randomforest":
+                classifier, train_err = \
+                    train_random_forest_regression(f_train, l_train, C)
+            error_test = []
+            error_test_baseline = []
+            for itest, fTest in enumerate(f_test):
+                R = regression_wrapper(classifier, method_name, fTest)
+                Rbaseline = np.mean(l_train)
+                error_test.append((R - l_test[itest]) *
+                                  (R - l_test[itest]))
+                error_test_baseline.append((Rbaseline - l_test[itest]) *
+                                           (Rbaseline - l_test[itest]))
+            error = np.array(error_test).mean()
+            error_baseline = np.array(error_test_baseline).mean()
+            errors.append(error)
+            errors_train.append(train_err)
+            errors_baseline.append(error_baseline)
+        errors_all.append(np.array(errors).mean())
+        er_train_all.append(np.array(errors_train).mean())
+        er_base_all.append(np.array(errors_baseline).mean())
 
     best_ind = np.argmin(errors_all)
 
@@ -805,7 +846,7 @@ def evaluate_regression(features, labels, n_exp, method_name, params):
                                                                 er_base_all[i]),
               end="")
         if i == best_ind:
-            print("\t\t best",end="")
+            print("\t\t best", end="")
         print("")
     return params[best_ind], errors_all[best_ind], er_base_all[best_ind]
 
@@ -868,7 +909,7 @@ def features_to_matrix(features):
 
 def pca_wrapper(features, dimensions):
     features, labels = features_to_matrix(features)
-    pca = sklearn.decomposition.PCA(n_components = dimensions)
+    pca = sklearn.decomposition.PCA(n_components=dimensions)
     pca.fit(features)
     coeff = pca.components_
     coeff = coeff[:, 0:dimensions]
@@ -924,7 +965,7 @@ def evaluate_model_for_folders(input_test_folders, model_name, model_type,
                                    "data/models/svm_rbf_4class",
                                    "svm_rbf", "speech")
     """
-    
+
     class_names = []
     y_true_binary = []
     y_true = []
@@ -945,7 +986,7 @@ def evaluate_model_for_folders(input_test_folders, model_name, model_type,
             c, p, probs_names = file_classification(w, model_name, model_type)
             y_pred.append(c)
             y_true.append(probs_names.index(class_names[i]))
-            if i==probs_names.index(positive_class):
+            if i == probs_names.index(positive_class):
                 y_true_binary.append(1)
             else:
                 y_true_binary.append(0)
@@ -954,7 +995,8 @@ def evaluate_model_for_folders(input_test_folders, model_name, model_type,
             probs_positive.append(prob_positive)
     pre, rec, thr_prre = sklearn.metrics.precision_recall_curve(y_true_binary,
                                                                 probs_positive)
-    fpr, tpr, thr_roc = sklearn.metrics.roc_curve(y_true_binary, probs_positive)
+    fpr, tpr, thr_roc = sklearn.metrics.roc_curve(
+        y_true_binary, probs_positive)
     cm = sklearn.metrics.confusion_matrix(y_true, y_pred)
     rec_c,  pre_c, f1_c = compute_class_rec_pre_f1(cm)
     f1 = (sklearn.metrics.f1_score(y_true, y_pred, average='macro'))
@@ -981,12 +1023,13 @@ def evaluate_model_for_folders(input_test_folders, model_name, model_type,
         mark_prop3 = dict(color='rgba(250, 150, 150, 0.5)',
                           line=dict(color='rgba(250, 150, 150, 1)', width=3))
         b1 = go.Bar(x=class_names, y=rec_c, name="Recall", marker=mark_prop1)
-        b2 = go.Bar(x=class_names, y=pre_c, name="Precision", marker=mark_prop2)
+        b2 = go.Bar(x=class_names, y=pre_c,
+                    name="Precision", marker=mark_prop2)
         b3 = go.Bar(x=class_names, y=f1_c, name="F1", marker=mark_prop3)
 
-        figs.append_trace(heatmap, 1, 1);
+        figs.append_trace(heatmap, 1, 1)
         figs.append_trace(b1, 1, 2)
-        figs.append_trace(b2, 1, 2);
+        figs.append_trace(b2, 1, 2)
         figs.append_trace(b3, 1, 2)
         figs.append_trace(go.Scatter(x=thr_prre, y=pre, name="Precision",
                                      marker=mark_prop1), 2, 1)
@@ -1099,7 +1142,8 @@ def file_regression(input_file, model_name, model_type):
                 or model_type == 'randomforest':
             model, mean, std, _, _, _, _, _ = load_model(r, True)
         curFV = (mid_features - mean) / std  # normalization
-        R.append(regression_wrapper(model, model_type, curFV))  # classification
+        # classification
+        R.append(regression_wrapper(model, model_type, curFV))
     return R, regression_names
 
 
@@ -1153,7 +1197,7 @@ def train_speaker_models():
                     for name in os.listdir(dir_name)
                     if os.path.isdir(os.path.join(dir_name, name))]
     extract_features_and_train(list_of_dirs, mt_win, mt_step, st_win, st_step,
-                    "knn", "data/knnSpeakerAll",
+                               "knn", "data/knnSpeakerAll",
                                compute_beat=False, train_percentage=0.50)
 
     dir_name = "DIARIZATION_ALL/female_male"
@@ -1161,7 +1205,7 @@ def train_speaker_models():
                     for name in os.listdir(dir_name)
                     if os.path.isdir(os.path.join(dir_name, name))]
     extract_features_and_train(list_of_dirs, mt_win, mt_step, st_win, st_step,
-                    "knn", "data/knnSpeakerFemaleMale",
+                               "knn", "data/knnSpeakerFemaleMale",
                                compute_beat=False, train_percentage=0.50)
 
 
